
LIBRARY        ARJCRYPT  INITGLOBAL
EXETYPE        OS2

DESCRIPTION    '32-bit GOST 28147-89 encryption plug-in for ARJ/2'
PROTMODE
DATA           MULTIPLE
CODE           LOADONCALL

EXPORTS
 entry         @1
